//
//    FILE: hp82165a.cpp
// VERSION: 0.1.0
// PURPOSE: HP-IL GPIO Interface for Arduino

#include "hp82165a.h" 

// <<constructor>>  
hp82165a::hp82165a(){

  // HandShake input
  pinMode(GETO_PIN, INPUT);
  pinMode(DAVO_PIN, INPUT);
  pinMode(RDYO_PIN, INPUT);
  pinMode(DACO_PIN, INPUT);

  // HandShake output
  pinMode(RDYI_PIN, OUTPUT);
  pinMode(DACI_PIN, OUTPUT);
  pinMode(DAVI_PIN, OUTPUT);
  // Detect interface push button
  pinMode(MSRQ_PIN, INPUT); 

  // Visual feedback
  pinMode(LED_PIN, OUTPUT);

  // Reset device
  clear();
}

// <<destructor>>
hp82165a::~hp82165a(){

}

// Handle interface signals
int hp82165a::watch(){

  int watch_result = WATCH;

/*
  unsigned int oldstate = state;

  bitWrite(handshake[1], 0, digitalRead(GETO_PIN));
  bitWrite(handshake[1], 1, digitalRead(DAVO_PIN));
  bitWrite(handshake[1], 2, digitalRead(MSRQ_PIN));
  
  if ( handshake[1] != handshake[0] ) {
    Serial.print("Handshake :");
    Serial.println(handshake[1]);
    handshake[0] = handshake[1];
  }
*/  
  switch (state) {

  case INIT:
  
    // Is interface ready ?
    if ( (digitalRead(DAVO_PIN) == HIGH) && 
         (digitalRead(GETO_PIN) == HIGH) && 
         (digitalRead(MSRQ_PIN) == HIGH) ) {

      // Ready to receive data
      state = IDLE;
      digitalWrite(RDYI_PIN,LOW);   // Accept data
      
    }
    break;

  case IDLE :

     /************************************************
     * IDLE State
     * - Start incoming data reception
     * - Check trigger signals
     * - Check service signals
     * - Send pending output
     ************************************************/

    // 1 - Does GPIO has data for us ?
    if (digitalRead(DAVO_PIN) == LOW) {

      // Device temporary not ready to receive data
      digitalWrite(RDYI_PIN, HIGH);

      // Input data into buffer
      listen();

      // Acknowledge data recept
      digitalWrite(DACI_PIN, LOW);

      // Visual feedback on (RX)   
      digitalWrite(LED_PIN, HIGH);

      // New command received ?
      if ( indata[inwrite] == LF ) {
        // Flush reply buffer if not sent, to make sure next available reply concern last command
        outread = outwrite;
        // Notify we have a valid command
        watch_result = COMMAND;
      }

      // Wait interface to leave the data
      timeout = millis();
      state = INWAIT;

    }

    // 2 - Trigger interrupt
    if ( digitalRead(GETO_PIN) == LOW ) {

      // Serial.println("Trigger Up");

      // Wait interface to finish trigger
      timeout = millis();
      state = TRGWAIT;
    
    }

    // 3 - Service interrupt
    if ( digitalRead(MSRQ_PIN) == LOW ) {

      //Serial.println("Service Up");
 
      // Event is fired immediately
      watch_result = SERVICE;
 
      // Wait interface to finish service
      timeout = millis();
      state = SRQWAIT;
      
    }

    // 4 - All inputs proceeded, do we have data to output back ?
    if (outread != outwrite) {

      // Interface ready to receive our data ?
      if (digitalRead(RDYO_PIN) == LOW) {

        // Visual feedback on (TX)   
        digitalWrite(LED_PIN, HIGH);

        // Device not ready to receive data
        digitalWrite(RDYI_PIN, HIGH);

        // Set bus A for output
        halfduplex(OUTPUT);

        // Output data from buffer
        talk();

        // Tell interface we have data ready for her
        digitalWrite(DAVI_PIN, LOW);

        // Wait interface to take the data
        timeout = millis();
        state = OUTWAIT;

      }

    }

    // end IDLE
    break;

  case INWAIT :

    /************************************************
     * INWAIT State
     * - Terminate GPIO data reception handshake
     ************************************************/

    // Interface finished with data ?
    if ( digitalRead(DAVO_PIN) == HIGH ) {

      // Device also done
      digitalWrite(DACI_PIN, HIGH);

      // Ready to get another byte
      digitalWrite(RDYI_PIN, LOW);
      state = IDLE;

      // Visual feedback off    
      digitalWrite(LED_PIN, LOW);

    } 
    else if ( (millis() - timeout) > PULSETMO )  {

      // DAVO time out
      // Serial.println( "DAVO time out.");
      service();
      watch_result = SERVICE;
      state = INIT;

      // Visual feedback off    
      digitalWrite(LED_PIN, LOW);

    }   
    break;  

  case TRGWAIT:

    /************************************************
     * TRGWAIT State
     * - Terminate GPIO trigger
     ************************************************/

    // Interface finished trigger ?
    if ( digitalRead(GETO_PIN) == HIGH ) {

      // Serial.println("Trigger Down");

      // Trigger event
      watch_result = TRIGGER;
      state = IDLE;

    } 
    else if ( (millis() - timeout) > PULSETMO )  {

      // GETO time out
      // Serial.println( "GETO time out.");
      service();
      watch_result = SERVICE;
      state = INIT;

    }   
    break;  

  case SRQWAIT:

    /************************************************
     * Service State
     * - Terminate GPIO service request
     ************************************************/

    // Interface finished service request ?
    if ( digitalRead(MSRQ_PIN) == HIGH || (millis() - timeout) > SRQTMO ) {

      //Serial.println("MSRQ Down");

      // Finish event
      state = INIT;

    } 
    break;  

  case OUTWAIT:

    /************************************************
     * OUTWAIT State
     * - Terminate GPIO data transmit handshake
     ************************************************/

    // Interface accepted data ?
    if ( digitalRead(DACO_PIN) == LOW ) {

      // Device also done
      digitalWrite(DAVI_PIN, HIGH);
      
      // Ready to get another byte
      halfduplex(INPUT);
      digitalWrite(RDYI_PIN, LOW);
      state = IDLE;

      // Visual feedback off    
      digitalWrite(LED_PIN, LOW);

    } 
    else if ( (millis() - timeout) > PULSETMO )  {

      // DACO time out
      // Serial.println( "DACO time out.");
      service();
      watch_result = SERVICE;
      state = INIT;

      // Visual feedback off    
      digitalWrite(LED_PIN, LOW);
    }   
    break;  

  } // end switch state

/*
  if (state != oldstate) {
    Serial.print("State change :");                  
    Serial.print(oldstate); 
    Serial.print(" -> ");                  
    Serial.println(state); 
    oldstate = state;    
  }
*/

  return watch_result;

}

/************************************************
 * GPIO received command
 ************************************************/
String hp82165a::enter(){

  String data = "";
  // String trace;

  if ( indata[inwrite] == LF ) {

    // Retrieve command line
    do {
      inread++;
      data += (char) indata[inread];
    } 
    while ( indata[inread] != LF );

    // Format command
    data.trim();
    data.toUpperCase();
 
    //  trace = "HP-IL GPIO Interface data Received = [";
    //  trace += data.trim();
    //  trace += "]";
    //  Serial.println(trace); 

  }   
  return data;  
}

/************************************************
 * GPIO fill output buffer with command to send
 ************************************************/
void hp82165a::output(char *command){

  //String strcmd = command;
  output((String) command);
  
}

void hp82165a::output(String command){

  String data;
  // String trace;

  // Format reply
  data = command;
  data.trim();
  data.toUpperCase();  

  // Add EOL (if not already present)
  if ( data.length() > 0 ) {
    if ( data[data.length()] != (char) LF ) {
      data += "\r\n";
    }
  }
  
  // Add reply to output buffer
  for (int i=0; i<data.length(); i++) {
    outwrite++;
    if (outwrite == outread) {

      // Output buffer is full
      // Serial.println("Output buffer is full...");   
      outwrite--;

      // Ask for MSRQ
      service();

    } 
    else {        
      outdata[outwrite] = data[i];
    }
  }

  //  trace = "HP-IL GPIO Interface data Sent = [";
  //  trace += data.trim();
  //  trace += "]";
  //  Serial.println(trace); 

}

/************************************************
 * GPIO Manual Service Request
 ************************************************/
void hp82165a::service(void) {

  // Serial.println("Service Request...");

  // Ask for MSRQ 
  pinMode(MSRQ_PIN, OUTPUT); 
  digitalWrite(MSRQ_PIN,LOW);
  // Leave enough time for controller periodic IDY frame
  delay(850); 
  digitalWrite(MSRQ_PIN,HIGH); 
  // As signal is high, pullup resistor is set when mode changed
  pinMode(MSRQ_PIN, INPUT); 

}

/************************************************
 * GPIO Interface clear
 ************************************************/
void hp82165a::clear(void) {

  unsigned int i;

  // Buffers (filled with LF command termination)
  for (i=0; i <= 0xFF; i++) {
    indata[i]=LF;
    outdata[i]=LF;   
  }

  // Buffers pointers (setup at the end)
  inwrite  = 0xFF;
  inread   = 0xFF;
  outwrite = 0xFF;
  outread  = 0xFF;

  // Data BUS ready to accept data
  halfduplex(INPUT);

  // Pull-up resistors on input pins to avoid unpredictable results when interface unplugged
  digitalWrite(GETO_PIN, HIGH);
  digitalWrite(DAVO_PIN, HIGH);
  digitalWrite(RDYO_PIN, HIGH);
  digitalWrite(DACO_PIN, HIGH);
  digitalWrite(MSRQ_PIN, HIGH);

  // HandShake uses negative logic (unless modified in registers) 
  digitalWrite(RDYI_PIN, HIGH);
  digitalWrite(DACI_PIN, HIGH);
  digitalWrite(DAVI_PIN, HIGH);

  // Hold GETO reverse logic on pin 2
  // attachInterrupt(0, trigger, CHANGE);

  // Default timeout start
  timeout = millis();

/*
  // Ready to receive data
  state = IDLE;
  digitalWrite(RDYI_PIN,LOW);   // Accept data
*/

  // Wait interface ready
  state = INIT;

}

/************************************************
 * GPIO Half Duplex Bus direction  
 ************************************************/
void hp82165a::halfduplex(unsigned int mode) {

  if (mode == INPUT) {

    // Disconnect pullup resistor 
    digitalWrite(DA0_PIN, LOW);
    digitalWrite(DA1_PIN, LOW);
    digitalWrite(DA2_PIN, LOW);
    digitalWrite(DA3_PIN, LOW);
    digitalWrite(DA4_PIN, LOW);
    digitalWrite(DA5_PIN, LOW);
    digitalWrite(DA6_PIN, LOW);
    digitalWrite(DA7_PIN, LOW);
  }

  // Data BUS A
  pinMode(DA0_PIN, mode);
  pinMode(DA1_PIN, mode);
  pinMode(DA2_PIN, mode);
  pinMode(DA3_PIN, mode);
  pinMode(DA4_PIN, mode);
  pinMode(DA5_PIN, mode);
  pinMode(DA6_PIN, mode);
  pinMode(DA7_PIN, mode);

}

/************************************************
 * GPIO Input data from interface to buffer
 ************************************************/
void hp82165a::listen (void) {

  inwrite++;
  if (inwrite == inread) {

    // Input buffer is full
    // Serial.println("Input buffer is full...");   
    inwrite--;

    // Ask for MSRQ
    service();

  } 
  else {        

    // Store input buffer
    indata[inwrite] = 0;
    bitWrite(indata[inwrite], 0, digitalRead(DA0_PIN));
    bitWrite(indata[inwrite], 1, digitalRead(DA1_PIN));
    bitWrite(indata[inwrite], 2, digitalRead(DA2_PIN));
    bitWrite(indata[inwrite], 3, digitalRead(DA3_PIN));
    bitWrite(indata[inwrite], 4, digitalRead(DA4_PIN));
    bitWrite(indata[inwrite], 5, digitalRead(DA5_PIN));
    bitWrite(indata[inwrite], 6, digitalRead(DA6_PIN));
    bitWrite(indata[inwrite], 7, digitalRead(DA7_PIN));  

    // Fake null characters may be generated at interface power up/down
    if ( indata[inwrite] == 0 ) {
      // Replace nulls by spaces
      indata[inwrite] = 0x20;
    }
    
    /*
    Serial.print("Listen [");   
    Serial.print(indata[inwrite]);   
    Serial.println("]");   
    */
    
  }

}

/************************************************
 * GPIO Output data from buffer to interface
 ************************************************/
void hp82165a::talk (void) {

  // Write data
  outread++;
  digitalWrite(DA0_PIN, bitRead(outdata[outread],0));
  digitalWrite(DA1_PIN, bitRead(outdata[outread],1));
  digitalWrite(DA2_PIN, bitRead(outdata[outread],2));
  digitalWrite(DA3_PIN, bitRead(outdata[outread],3));
  digitalWrite(DA4_PIN, bitRead(outdata[outread],4));
  digitalWrite(DA5_PIN, bitRead(outdata[outread],5));
  digitalWrite(DA6_PIN, bitRead(outdata[outread],6));
  digitalWrite(DA7_PIN, bitRead(outdata[outread],7));

}
