//
//    FILE: hp82165a.h
// VERSION: 0.1.0
// PURPOSE: HP-IL GPIO Interface for Arduino

#ifndef hp82165a_H
#define hp82165a_H

#include <Arduino.h> 

// GPIO Data A
const int DA0_PIN  = 22;    
const int DA1_PIN  = 24;    
const int DA2_PIN  = 26;    
const int DA3_PIN  = 28;    
const int DA4_PIN  = 30;    
const int DA5_PIN  = 32;    
const int DA6_PIN  = 34;  
const int DA7_PIN  = 36;  

// HandShake from interface
const int GETO_PIN = 23;
const int DAVO_PIN = 25; 
const int RDYO_PIN = 27;
const int DACO_PIN = 29;

// HandShake to interface
const int RDYI_PIN = 31;
const int DACI_PIN = 33;
const int DAVI_PIN = 35;
const int MSRQ_PIN = 37;

// Visual feedback 
const int LED_PIN     = 13;    // internal led   

// Device states
const int INIT    = 0x00;
const int IDLE    = 0x01;
const int INWAIT  = 0x02;
const int OUTWAIT = 0x04;
const int TRGWAIT = 0x08;
const int SRQWAIT = 0x10;

// Watch results
const int WATCH    = 0x00;
const int TRIGGER  = 0x01;
const int COMMAND  = 0x03;
const int SERVICE  = 0x05;

// GPIO timeout (ms) for DAVO & DACO Pulse width
const int PULSETMO = 10;   // Higher than 5 ms
const int SRQTMO   = 1000; // Higher than 1s

// End of line
const byte CR = 0x0D;
const byte LF = 0x0A;

class hp82165a{
  
private:

  // Device state
  unsigned int  state;
  unsigned long timeout;
  byte handshake[2];

  // GPIO Input buffer
  byte indata[256];
  byte inwrite;
  byte inread;

  // GPIO Output buffer
  byte outdata[256];
  byte outwrite;
  byte outread;
  
  // private methods
  void clear(void);
  void halfduplex(unsigned int mode);
  void listen (void);
  void talk (void);
 
public:

  // public methods
  hp82165a();
  ~hp82165a();
  int watch();
  String enter();
  void output(String command);
  void output(char *command);
  void service();

};

#endif

